/*
  This file is part of TALER
  Copyright (C) 2014-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_post-reserves-RESERVE_PUB-close.c
 * @brief Implementation of the POST /reserves/$RESERVE_PUB/close requests
 * @author Christian Grothoff
 */
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP close codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_json_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler/taler_json_lib.h"
#include "exchange_api_handle.h"
#include "taler/taler_signatures.h"
#include "exchange_api_curl_defaults.h"


/**
 * @brief A POST /reserves/$RID/close Handle
 */
struct TALER_EXCHANGE_PostReservesCloseHandle
{

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Base URL of the exchange.
   */
  char *base_url;

  /**
   * The url for this request, set during _start.
   */
  char *url;

  /**
   * Context for #TEH_curl_easy_post(). Keeps the data that must
   * persist for Curl to make the upload.
   */
  struct TALER_CURL_PostContext post_ctx;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_PostReservesCloseCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_EXCHANGE_POST_RESERVES_CLOSE_RESULT_CLOSURE *cb_cls;

  /**
   * Private key of the reserve.
   */
  struct TALER_ReservePrivateKeyP reserve_priv;

  /**
   * Public key of the reserve we are querying.
   */
  struct TALER_ReservePublicKeyP reserve_pub;

  /**
   * Target payto URI to send the reserve balance to (optional).
   */
  struct TALER_FullPayto target_payto_uri;

  /**
   * When did we make the request.
   */
  struct GNUNET_TIME_Timestamp ts;

};


/**
 * We received an #MHD_HTTP_OK close code. Handle the JSON
 * response.
 *
 * @param prch handle of the request
 * @param j JSON response
 * @return #GNUNET_OK on success
 */
static enum GNUNET_GenericReturnValue
handle_reserves_close_ok (struct TALER_EXCHANGE_PostReservesCloseHandle *prch,
                          const json_t *j)
{
  struct TALER_EXCHANGE_PostReservesCloseResponse rs = {
    .hr.reply = j,
    .hr.http_status = MHD_HTTP_OK,
  };
  struct GNUNET_JSON_Specification spec[] = {
    TALER_JSON_spec_amount_any ("wire_amount",
                                &rs.details.ok.wire_amount),
    GNUNET_JSON_spec_end ()
  };

  if (GNUNET_OK !=
      GNUNET_JSON_parse (j,
                         spec,
                         NULL,
                         NULL))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  prch->cb (prch->cb_cls,
            &rs);
  prch->cb = NULL;
  GNUNET_JSON_parse_free (spec);
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP /reserves/$RID/close request.
 *
 * @param cls the `struct TALER_EXCHANGE_PostReservesCloseHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_reserves_close_finished (void *cls,
                                long response_code,
                                const void *response)
{
  struct TALER_EXCHANGE_PostReservesCloseHandle *prch = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_PostReservesCloseResponse rs = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  prch->job = NULL;
  switch (response_code)
  {
  case 0:
    rs.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    if (GNUNET_OK !=
        handle_reserves_close_ok (prch,
                                  j))
    {
      GNUNET_break_op (0);
      rs.hr.http_status = 0;
      rs.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
    }
    break;
  case MHD_HTTP_BAD_REQUEST:
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    GNUNET_break (0);
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_FORBIDDEN:
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    GNUNET_break (0);
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_NOT_FOUND:
    /* Nothing really to verify, this should never
       happen, we should pass the JSON reply to the application */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_CONFLICT:
    /* Insufficient balance to inquire for reserve close */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_UNAVAILABLE_FOR_LEGAL_REASONS:
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    if (GNUNET_OK !=
        TALER_EXCHANGE_parse_451 (&rs.details.unavailable_for_legal_reasons,
                                  j))
    {
      GNUNET_break_op (0);
      rs.hr.http_status = 0;
      rs.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
    }
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  default:
    /* unexpected response code */
    GNUNET_break_op (0);
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for reserves close\n",
                (unsigned int) response_code,
                (int) rs.hr.ec);
    break;
  }
  if (NULL != prch->cb)
  {
    prch->cb (prch->cb_cls,
              &rs);
    prch->cb = NULL;
  }
  TALER_EXCHANGE_post_reserves_close_cancel (prch);
}


struct TALER_EXCHANGE_PostReservesCloseHandle *
TALER_EXCHANGE_post_reserves_close_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const struct TALER_ReservePrivateKeyP *reserve_priv)
{
  struct TALER_EXCHANGE_PostReservesCloseHandle *prch;

  prch = GNUNET_new (struct TALER_EXCHANGE_PostReservesCloseHandle);
  prch->ctx = ctx;
  prch->base_url = GNUNET_strdup (url);
  prch->reserve_priv = *reserve_priv;
  GNUNET_CRYPTO_eddsa_key_get_public (&reserve_priv->eddsa_priv,
                                      &prch->reserve_pub.eddsa_pub);
  prch->target_payto_uri.full_payto = NULL;
  return prch;
}


enum GNUNET_GenericReturnValue
TALER_EXCHANGE_post_reserves_close_set_options_ (
  struct TALER_EXCHANGE_PostReservesCloseHandle *prch,
  unsigned int num_options,
  const struct TALER_EXCHANGE_PostReservesCloseOptionValue options[])
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    const struct TALER_EXCHANGE_PostReservesCloseOptionValue *opt = &options[i];

    switch (opt->option)
    {
    case TALER_EXCHANGE_POST_RESERVES_CLOSE_OPTION_END:
      return GNUNET_OK;
    case TALER_EXCHANGE_POST_RESERVES_CLOSE_OPTION_PAYTO_URI:
      GNUNET_free (prch->target_payto_uri.full_payto);
      prch->target_payto_uri.full_payto
        = GNUNET_strdup (opt->details.payto_uri.full_payto);
      break;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_EXCHANGE_post_reserves_close_start (
  struct TALER_EXCHANGE_PostReservesCloseHandle *prch,
  TALER_EXCHANGE_PostReservesCloseCallback cb,
  TALER_EXCHANGE_POST_RESERVES_CLOSE_RESULT_CLOSURE *cb_cls)
{
  CURL *eh;
  char arg_str[sizeof (struct TALER_ReservePublicKeyP) * 2 + 32];
  struct TALER_ReserveSignatureP reserve_sig;
  struct TALER_FullPaytoHashP h_payto;
  json_t *close_obj;

  prch->cb = cb;
  prch->cb_cls = cb_cls;
  prch->ts = GNUNET_TIME_timestamp_get ();
  {
    char pub_str[sizeof (struct TALER_ReservePublicKeyP) * 2];
    char *end;

    end = GNUNET_STRINGS_data_to_string (
      &prch->reserve_pub,
      sizeof (prch->reserve_pub),
      pub_str,
      sizeof (pub_str));
    *end = '\0';
    GNUNET_snprintf (arg_str,
                     sizeof (arg_str),
                     "reserves/%s/close",
                     pub_str);
  }
  prch->url = TALER_url_join (prch->base_url,
                              arg_str,
                              NULL);
  if (NULL == prch->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  if (NULL != prch->target_payto_uri.full_payto)
    TALER_full_payto_hash (prch->target_payto_uri,
                           &h_payto);
  TALER_wallet_reserve_close_sign (prch->ts,
                                   (NULL != prch->target_payto_uri.full_payto)
                                   ? &h_payto
                                   : NULL,
                                   &prch->reserve_priv,
                                   &reserve_sig);
  close_obj = GNUNET_JSON_PACK (
    GNUNET_JSON_pack_allow_null (
      TALER_JSON_pack_full_payto ("payto_uri",
                                  prch->target_payto_uri)),
    GNUNET_JSON_pack_timestamp ("request_timestamp",
                                prch->ts),
    GNUNET_JSON_pack_data_auto ("reserve_sig",
                                &reserve_sig));
  eh = TALER_EXCHANGE_curl_easy_get_ (prch->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&prch->post_ctx,
                              eh,
                              close_obj)) )
  {
    GNUNET_break (0);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    json_decref (close_obj);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  json_decref (close_obj);
  prch->job = GNUNET_CURL_job_add2 (prch->ctx,
                                    eh,
                                    prch->post_ctx.headers,
                                    &handle_reserves_close_finished,
                                    prch);
  if (NULL == prch->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_post_reserves_close_cancel (
  struct TALER_EXCHANGE_PostReservesCloseHandle *prch)
{
  if (NULL != prch->job)
  {
    GNUNET_CURL_job_cancel (prch->job);
    prch->job = NULL;
  }
  TALER_curl_easy_post_finished (&prch->post_ctx);
  GNUNET_free (prch->url);
  GNUNET_free (prch->base_url);
  GNUNET_free (prch->target_payto_uri.full_payto);
  GNUNET_free (prch);
}


/* end of exchange_api_post-reserves-RESERVE_PUB-close.c */
